from collections import OrderedDict
from typing import Any, Dict, List, Optional

import numpy as np

from irspack.definitions import InteractionMatrix
from irspack.evaluator.evaluator import METRIC_NAMES, EvaluatorCore, Metrics
from irspack.user_cold_start.recommenders import base
from irspack.utils.threading import get_n_threads


class UserColdStartEvaluator:
    def __init__(
        self,
        X: InteractionMatrix,
        profiles: base.ProfileMatrix,
        mb_size: int = 1024,
        n_threads: Optional[int] = None,
        cutoff: int = 20,
    ):
        assert X.shape[0] == profiles.shape[0]

        self.core = EvaluatorCore(X.astype(np.float64), [])
        self.profiles = profiles
        self.n_users = X.shape[0]
        self.n_items = X.shape[1]
        self.dim_profile = profiles.shape[1]
        self.mb_size = mb_size
        self.n_threads = get_n_threads(n_threads)
        self.cutoff = cutoff

    def get_score(self, model: base.BaseUserColdStartRecommender) -> Dict[str, Any]:
        metric_base = Metrics(self.n_items)
        for start in range(0, self.n_users, self.mb_size):
            end = min(start + self.mb_size, self.n_users)
            score_mb = model.get_score(self.profiles[start:end])
            metric = self.core.get_metrics(
                score_mb, self.cutoff, start, self.n_threads, False
            )
            metric_base.merge(metric)
        return metric_base.as_dict()

    def get_scores(
        self, model: base.BaseUserColdStartRecommender, cutoffs: List[int]
    ) -> Dict[str, float]:
        result: Dict[str, float] = OrderedDict()
        scores = self.get_scores_as_list(model, cutoffs)
        for cutoff, score in zip(cutoffs, scores):
            for metric_name in METRIC_NAMES:
                result[f"{metric_name}@{cutoff}"] = score[metric_name]
        return result

    def get_scores_as_list(
        self, model: base.BaseUserColdStartRecommender, cutoffs: List[int]
    ) -> List[Dict[str, float]]:
        n_items = model.n_items
        metrics: List[Metrics] = []
        for c in cutoffs:
            metrics.append(Metrics(n_items))
        n_validated = self.n_users
        block_end = n_validated
        mb_size = self.mb_size

        for chunk_start in range(0, block_end, mb_size):
            chunk_end = min(chunk_start + mb_size, block_end)
            score_mb = model.get_score(self.profiles[chunk_start:chunk_end])
            for i, cutoff in enumerate(cutoffs):
                chunked_metric = self.core.get_metrics(
                    score_mb, cutoff, chunk_start, self.n_threads, False
                )
                metrics[i].merge(chunked_metric)
        return [item.as_dict() for item in metrics]



from typing import Optional

import numpy as np

from ..definitions import DenseScoreArray, InteractionMatrix, UserIndexArray
from .base import BaseRecommender


class TopPopRecommender(BaseRecommender):
    """A simple recommender system based on the popularity of the items in the
    training set (without any personalization).

    Args:
        X_train Union[scipy.sparse.csr_matrix, scipy.sparse.csc_matrix]):
            Input interaction matrix.
    """

    score_: Optional[np.ndarray]

    def __init__(self, X_train: InteractionMatrix):

        super().__init__(X_train)
        self.score_ = None

    def _learn(self) -> None:
        self.score_ = self.X_train_all.sum(axis=0).astype(np.float64)

    @property
    def score(self) -> np.ndarray:
        if self.score_ is None:
            raise RuntimeError("The method called before ``learn``.")
        return self.score_

    def get_score(self, user_indices: UserIndexArray) -> DenseScoreArray:
        n_users: int = user_indices.shape[0]
        return np.repeat(self.score, n_users, axis=0).A

    def get_score_cold_user(self, X: InteractionMatrix) -> DenseScoreArray:
        n_users: int = X.shape[0]
        return np.repeat(self.score, n_users, axis=0).A



from abc import abstractmethod
from typing import Optional, Union

from irspack.definitions import InteractionMatrix
from irspack.recommenders._knn import (
    AsymmetricSimilarityComputer,
    CosineSimilarityComputer,
    JaccardSimilarityComputer,
    TverskyIndexComputer,
)
from irspack.recommenders.base import BaseUserSimilarityRecommender
from irspack.recommenders.knn import FeatureWeightingScheme
from irspack.utils import (
    get_n_threads,
    okapi_BM_25_weight,
    remove_diagonal,
    tf_idf_weight,
)


class BaseUserKNNRecommender(BaseUserSimilarityRecommender):
    def __init__(
        self,
        X_train_all: InteractionMatrix,
        shrinkage: float = 0.0,
        top_k: int = 100,
        n_threads: Optional[int] = None,
        feature_weighting: str = "NONE",
        bm25_k1: float = 1.2,
        bm25_b: float = 0.75,
    ):
        super().__init__(X_train_all)
        self.shrinkage = shrinkage
        self.top_k = top_k
        self.feature_weighting = FeatureWeightingScheme(feature_weighting)
        self.bm25_k1 = bm25_k1
        self.bm25_b = bm25_b
        self.n_threads = get_n_threads(n_threads)

    @abstractmethod
    def _create_computer(
        self, X: InteractionMatrix
    ) -> Union[
        CosineSimilarityComputer,
        AsymmetricSimilarityComputer,
        JaccardSimilarityComputer,
        TverskyIndexComputer,
    ]:
        raise NotImplementedError("")

    def _learn(self) -> None:
        if self.feature_weighting == FeatureWeightingScheme.NONE:
            X_weighted = self.X_train_all
        elif self.feature_weighting == FeatureWeightingScheme.TF_IDF:
            X_weighted = tf_idf_weight(self.X_train_all)
        elif self.feature_weighting == FeatureWeightingScheme.BM_25:
            X_weighted = okapi_BM_25_weight(self.X_train_all, self.bm25_k1, self.bm25_b)
        else:
            raise RuntimeError("Unknown weighting scheme.")

        computer = self._create_computer(X_weighted)
        self.U_ = remove_diagonal(
            computer.compute_similarity(self.X_train_all, self.top_k)
        )


class CosineUserKNNRecommender(BaseUserKNNRecommender):
    r"""K-nearest neighbor recommender system based on cosine similarity. That is, the similarity matrix ``U`` is given by (row-wise top-k restricted)

    .. math::

        \mathrm{U}_{u,v} = \begin{cases}
            \frac{\sum_{i} X_{ui} X_{vi}}{||X_{u*}||_2 ||X_{v*}||_2 + \mathrm{shrinkage}} & (\text{if normalize = True}) \\
            \sum_{i} X_{ui} X_{vi} & (\text{if normalize = False})
        \end{cases}


    Args:
        X_train_all (Union[scipy.sparse.csr_matrix, scipy.sparse.csc_matrix]):
            Input interaction matrix.
        shrinkage (float, optional):
            The shrinkage parameter for regularization. Defaults to 0.0.
        normalize (bool, optional):
            Whether to normalize the similarity. Defaults to False.
        top_k (int, optional):
            Specifies the maximal number of allowed neighbors. Defaults to 100.
        feature_weighting (str, optional):
            Specifies how to weight the feature. Must be one of:

                - "NONE" : no feature weighting
                - "TF_IDF" : TF-IDF weighting
                - "BM_25" : `Okapi BM-25 weighting <https://en.wikipedia.org/wiki/Okapi_BM25>`_

            Defaults to "NONE".
        bm25_k1 (float, optional):
            The k1 parameter for BM25. Ignored if ``feature_weighting`` is not "BM_25". Defaults to 1.2.
        bm25_b (float, optional):
            The b parameter for BM25. Ignored if ``feature_weighting`` is not "BM_25". Defaults to 0.75.
        n_threads (Optional[int], optional): Specifies the number of threads to use for the computation.
            If ``None``, the environment variable ``"IRSPACK_NUM_THREADS_DEFAULT"`` will be looked up,
            and if there is no such an environment variable, it will be set to ``os.cpu_count()``. Defaults to None.
    """

    def __init__(
        self,
        X_train_all: InteractionMatrix,
        shrinkage: float = 0.0,
        normalize: bool = True,
        top_k: int = 100,
        feature_weighting: str = "NONE",
        bm25_k1: float = 1.2,
        bm25_b: float = 0.75,
        n_threads: Optional[int] = None,
    ):
        super().__init__(
            X_train_all,
            shrinkage,
            top_k,
            n_threads,
            feature_weighting=feature_weighting,
            bm25_k1=bm25_k1,
            bm25_b=bm25_b,
        )
        self.normalize = normalize

    def _create_computer(self, X: InteractionMatrix) -> CosineSimilarityComputer:
        return CosineSimilarityComputer(
            X, self.shrinkage, self.normalize, self.n_threads
        )


class AsymmetricCosineUserKNNRecommender(BaseUserKNNRecommender):
    r"""K-nearest neighbor recommender system based on asymmetric cosine similarity. That is, the similarity matrix ``U`` is given by (row-wise top-k restricted)

    .. math::

        \mathrm{U}_{u,v} = \frac{\sum_{i} X_{ui} X_{vi}}{||X_{u*}||^{2\alpha}_2 ||X_{v*}||^{2(1-\alpha)}_2 + \mathrm{shrinkage}}

    Args:
        X_train_all (Union[scipy.sparse.csr_matrix, scipy.sparse.csc_matrix]):
            Input interaction matrix.
        shrinkage (float, optional):
            The shrinkage parameter for regularization. Defaults to 0.0.
        alpha (bool, optional):
            Specifies :math:`\alpha`. Defaults to 0.5.
        top_k (int, optional):
            Specifies the maximal number of allowed neighbors. Defaults to 100.
        feature_weighting (str, optional):
            Specifies how to weight the feature. Must be one of:

                - "NONE" : no feature weighting
                - "TF_IDF" : TF-IDF weighting
                - "BM_25" : `Okapi BM-25 weighting <https://en.wikipedia.org/wiki/Okapi_BM25>`_

            Defaults to "NONE".
        bm25_k1 (float, optional):
            The k1 parameter for BM25. Ignored if ``feature_weighting`` is not "BM_25". Defaults to 1.2.
        bm25_b (float, optional):
            The b parameter for BM25. Ignored if ``feature_weighting`` is not "BM_25". Defaults to 0.75.
        n_threads (Optional[int], optional): Specifies the number of threads to use for the computation.
            If ``None``, the environment variable ``"IRSPACK_NUM_THREADS_DEFAULT"`` will be looked up,
            and if there is no such an environment variable, it will be set to ``os.cpu_count()``. Defaults to None.
    """

    def __init__(
        self,
        X_train_all: InteractionMatrix,
        shrinkage: float = 0.0,
        alpha: float = 0.5,
        top_k: int = 100,
        feature_weighting: str = "NONE",
        bm25_k1: float = 1.2,
        bm25_b: float = 0.75,
        n_threads: Optional[int] = None,
    ):
        super().__init__(
            X_train_all,
            shrinkage,
            top_k,
            n_threads,
            feature_weighting=feature_weighting,
            bm25_k1=bm25_k1,
            bm25_b=bm25_b,
        )
        self.alpha = alpha

    def _create_computer(self, X: InteractionMatrix) -> AsymmetricSimilarityComputer:
        return AsymmetricSimilarityComputer(
            X, self.shrinkage, self.alpha, self.n_threads
        )



from irspack.user_cold_start.optimizers import (
    BaseOptimizer as BaseUserColdStartOptimizer,
)
from irspack.user_cold_start.optimizers import (
    LinearMethodOptimizer,
    TopPopularOptimizer,
    UserCBCosineKNNOptimizer,
)
from irspack.user_cold_start.recommenders import (
    BaseUserColdStartRecommender,
    LinearMethodRecommender,
    TopPopularRecommender,
    UserCBCosineKNNRecommender,
)

from .evaluator import UserColdStartEvaluator

__all__ = [
    "BaseUserColdStartRecommender",
    "UserColdStartEvaluator",
    "LinearMethodRecommender",
    "TopPopularRecommender",
    "UserCBCosineKNNRecommender",
    "UserCBCosineKNNOptimizer",
    "TopPopularOptimizer",
    "LinearMethodOptimizer",
]
try:
    from .cb2cf import CB2IALSOptimizer, CB2TruncatedSVDOptimizer

    __all__.extend(["CB2IALSOptimizer", "CB2TruncatedSVDOptimizer"])
    try:
        from .cb2cf import CB2BPRFMOptimizer

        __all__.append("CB2BPRFMOptimizer")
    except:
        pass
except:
    # must be jax-related error
    pass



from ._core import EvaluatorCore, Metrics
from .evaluator import METRIC_NAMES, Evaluator, EvaluatorWithColdUser, TargetMetric

__all__ = [
    "Evaluator",
    "Metrics",
    "TargetMetric",
    "METRIC_NAMES",
    "EvaluatorCore",
    "EvaluatorWithColdUser",
]



import logging
import re
import time
from abc import ABCMeta
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, no_type_check

import optuna
import pandas as pd

from irspack.utils.default_logger import get_default_logger

from ..evaluator import Evaluator
from ..parameter_tuning import Suggestion, is_valid_param_name, overwrite_suggestions
from ..recommenders.base import BaseRecommender, InteractionMatrix
from ..recommenders.base_earlystop import BaseRecommenderWithEarlyStopping


class LowMemoryError(RuntimeError):
    pass


_BaseOptimizerArgsString = """Args:
    data (Union[scipy.sparse.csr_matrix, scipy.sparse.csc_matrix]):
        The train data.
    val_evaluator (Evaluator):
        The validation evaluator which measures the performance of the recommenders.
    logger (Optional[logging.Logger], optional) :
        The logger used during the optimization steps. Defaults to None.
        If ``None``, the default logger of irspack will be used.
    suggest_overwrite (List[Suggestion], optional) :
        Customizes (e.g. enlarging the parameter region or adding new parameters to be tuned)
        the default parameter search space defined by ``default_tune_range``
        Defaults to list().
    fixed_params (Dict[str, Any], optional):
        Fixed parameters passed to recommenders during the optimization procedure.
        If such a parameter exists in ``default_tune_range``, it will not be tuned.
        Defaults to dict().
"""

_BaseOptimizerWithEarlyStoppingArgsString = """Args:
    data (Union[scipy.sparse.csr_matrix, scipy.sparse.csc_matrix]):
        The train data.
    val_evaluator (Evaluator):
        The validation evaluator which measures the performance of the recommenders.
    logger (Optional[logging.Logger], optional):
        The logger used during the optimization steps. Defaults to None.
        If ``None``, the default logger of irspack will be used.
    suggest_overwrite (List[Suggestion], optional):
        Customizes (e.g. enlarging the parameter region or adding new parameters to be tuned)
        the default parameter search space defined by ``default_tune_range``
        Defaults to list().
    fixed_params (Dict[str, Any], optional):
        Fixed parameters passed to recommenders during the optimization procedure.
        If such a parameter exists in ``default_tune_range``, it will not be tuned.
        Defaults to dict().
    max_epoch (int, optional):
        The maximal number of epochs for the training. Defaults to 512.
    validate_epoch (int, optional):
        The frequency of validation score measurement. Defaults to 5.
    score_degradation_max (int, optional):
        Maximal number of allowed score degradation. Defaults to 5. Defaults to 5.
"""


def optimizer_docstring(
    default_tune_range: Optional[List[Suggestion]],
    recommender_class: Optional[Type[BaseRecommender]],
) -> Optional[str]:
    if recommender_class is None:
        return None
    assert default_tune_range is not None
    if default_tune_range:

        ranges = ""
        for suggest in default_tune_range:
            ranges += f"          - ``{suggest!r}``\n"

        ranges += "\n"

    if issubclass(recommender_class, BaseRecommenderWithEarlyStopping):
        args_docstring = _BaseOptimizerWithEarlyStoppingArgsString
    else:
        args_docstring = _BaseOptimizerArgsString

    if default_tune_range:
        tune_range = f"""The default tune range is

{ranges}"""
    else:
        tune_range = "   There is no tunable parameters."
    docs = f"""Optimizer class for :class:`irspack.recommenders.{recommender_class.__name__}`.

{tune_range}

{args_docstring}

    """
    return docs


class OptimizerMeta(ABCMeta):
    optimizer_name_vs_optimizer_class: Dict[str, "OptimizerMeta"] = {}

    @no_type_check
    def __new__(
        mcs,
        name,
        bases,
        namespace,
        **kwargs,
    ):

        recommender_class: Optional[Type[BaseRecommender]] = namespace.get(
            "recommender_class"
        )
        default_tune_range: Optional[List[Suggestion]] = namespace.get(
            "default_tune_range"
        )
        if default_tune_range is None:
            assert recommender_class is None
        cls = super().__new__(mcs, name, bases, namespace, **kwargs)
        docs = optimizer_docstring(default_tune_range, recommender_class)
        if docs is not None:
            cls.__doc__ = docs
        mcs.optimizer_name_vs_optimizer_class[name] = cls
        return cls


def add_score_to_trial(
    trial: optuna.Trial, score: Dict[str, float], cutoff: int
) -> None:
    score_history: List[Tuple[int, Dict[str, float]]] = trial.study.user_attrs.get(
        "scores", []
    )
    score_history.append(
        (
            trial.number,
            {
                f"{score_name}@{cutoff}": score_value
                for score_name, score_value in score.items()
            },
        )
    )
    trial.study.set_user_attr("scores", score_history)


def study_to_dataframe(study: optuna.Study) -> pd.DataFrame:
    result_df: pd.DataFrame = study.trials_dataframe().set_index("number")

    # remove prefix
    result_df.columns = [
        re.sub(r"^(user_attrs|params)_", "", colname) for colname in result_df.columns
    ]

    trial_and_scores: List[Tuple[float, Dict[str, float]]] = study.user_attrs.get(
        "scores", []
    )
    score_df = pd.DataFrame(
        [x[1] for x in trial_and_scores],
        index=[x[0] for x in trial_and_scores],
    )
    score_df.index.name = "number"
    result_df = result_df.join(score_df, how="left")
    return result_df


class BaseOptimizer(object, metaclass=OptimizerMeta):

    recommender_class: Type[BaseRecommender]
    default_tune_range: List[Suggestion] = []

    @classmethod
    def tune_range_given_memory_budget(
        cls, X: InteractionMatrix, memory_in_mb: int
    ) -> List[Suggestion]:
        return []

    def __init__(
        self,
        data: InteractionMatrix,
        val_evaluator: Evaluator,
        logger: Optional[logging.Logger] = None,
        suggest_overwrite: List[Suggestion] = list(),
        fixed_params: Dict[str, Any] = dict(),
    ):

        if logger is None:
            logger = get_default_logger()

        self.logger = logger
        self._data = data
        self.val_evaluator = val_evaluator

        self.current_trial: int = 0
        self.best_val = float("inf")

        self.suggestions = overwrite_suggestions(
            self.default_tune_range, suggest_overwrite, fixed_params
        )
        self.fixed_params = fixed_params

    def _suggest(self, trial: optuna.Trial, prefix: str = "") -> Dict[str, Any]:
        parameters: Dict[str, Any] = dict()
        for s in self.suggestions:
            parameters[s.name] = s.suggest(trial, prefix)
        return parameters

    def get_model_arguments(
        self, *args: Any, **kwargs: Any
    ) -> Tuple[Tuple[Any, ...], Dict[str, Any]]:
        return args, kwargs

    def objective_function(
        self, param_prefix: str = ""
    ) -> Callable[[optuna.Trial], float]:
        """Returns the objective function that can be passed to ``optuna.Study`` .

        Returns:
            A callable that receives ``otpuna.Trial`` and returns float (like ndcg score).
        """

        def objective_func(trial: optuna.Trial) -> float:
            start = time.time()
            params = dict(**self._suggest(trial, param_prefix), **self.fixed_params)
            self.logger.info("Trial %s:", trial.number)
            self.logger.info("parameter = %s", params)

            arg, parameters = self.get_model_arguments(**params)

            recommender = self.recommender_class(self._data, *arg, **parameters)
            recommender.learn_with_optimizer(self.val_evaluator, trial)

            score = self.val_evaluator.get_score(recommender)
            end = time.time()

            time_spent = end - start
            self.logger.info(
                "Config %d obtained the following scores: %s within %f seconds.",
                trial.number,
                score,
                time_spent,
            )
            val_score = score[self.val_evaluator.target_metric.name]

            # max_epoch will be stored in learnt_config
            for param_name, param_val in recommender.learnt_config.items():
                trial.set_user_attr(param_name, param_val)

            add_score_to_trial(trial, score, self.val_evaluator.cutoff)

            return -val_score

        return objective_func

    def optimize_with_study(
        self,
        study: optuna.Study,
        n_trials: int = 20,
        timeout: Optional[int] = None,
    ) -> Tuple[Dict[str, Any], pd.DataFrame]:
        """Perform the optimization step using the user-created ``optuna.Study`` object.
        Creating and managing the study object will be convenient e.g. when you

            1. want to `store/resume the study using RDB backend <https://optuna.readthedocs.io/en/stable/tutorial/003_rdb.html>`_.
            2. want perform a `distributed optimization <https://optuna.readthedocs.io/en/stable/tutorial/004_distributed.html>`_.

        Args:
            study:
                The study object.
            n_trials:
                The number of expected trials (include pruned trial.). Defaults to 20.
            timeout:
                If set to some value (in seconds), the study will exit after that time period.
                Note that the running trials is not interrupted, though. Defaults to None.

        Returns:
            A tuple that consists of

                1. A dict containing the best paramaters.
                   This dict can be passed to the recommender as ``**kwargs``.
                2. A ``pandas.DataFrame`` that contains the history of optimization.

        """

        objective_func = self.objective_function()

        self.logger.info(
            """Start parameter search for %s over the range: %s""",
            type(self).recommender_class.__name__,
            self.suggestions,
        )

        study.optimize(objective_func, n_trials=n_trials, timeout=timeout)
        best_params = dict(
            **study.best_trial.params,
            **{
                key: val
                for key, val in study.best_trial.user_attrs.items()
                if is_valid_param_name(key)
            },
        )

        return best_params, study_to_dataframe(study)

    def optimize(
        self,
        n_trials: int = 20,
        timeout: Optional[int] = None,
        random_seed: Optional[int] = None,
    ) -> Tuple[Dict[str, Any], pd.DataFrame]:
        """Perform the optimization step.
        ``optuna.Study`` object is created inside this function.

        Args:
            n_trials:
                The number of expected trials (include pruned trial.). Defaults to 20.
            timeout:
                If set to some value (in seconds), the study will exit after that time period.
                Note that the running trials is not interrupted, though. Defaults to None.
            random_seed:
                The random seed to control ``optuna.samplers.TPESampler``. Defaults to None.

        Returns:
            A tuple that consists of

                1. A dict containing the best paramaters.
                   This dict can be passed to the recommender as ``**kwargs``.
                2. A ``pandas.DataFrame`` that contains the history of optimization.

        """
        study = optuna.create_study(
            sampler=optuna.samplers.TPESampler(seed=random_seed)
        )
        return self.optimize_with_study(study, n_trials, timeout)


class BaseOptimizerWithEarlyStopping(BaseOptimizer):
    """The Base Optimizer class for early-stoppable recommenders.

    The child class must define

        - ``recommender_class``
        - ``default_tune_range``


    Args:
        data (Union[scipy.sparse.csr_matrix, scipy.sparse.csc_matrix]):
            The train data.
        val_evaluator (Evaluator):
            The validation evaluator which measures the performance of the recommenders.
        logger (Optional[logging.Logger], optional):
            The logger used during the optimization steps. Defaults to None.
            If ``None``, the default logger of irspack will be used.
        suggest_overwrite (List[Suggestion], optional):
            Customizes (e.g. enlarging the parameter region or adding new parameters to be tuned)
            the default parameter search space defined by ``default_tune_range``
            Defaults to list().
        fixed_params (Dict[str, Any], optional):
            Fixed parameters passed to recommenders during the optimization procedure.
            If such a parameter exists in ``default_tune_range``, it will not be tuned.
            Defaults to dict().
        max_epoch (int, optional):
            The maximal number of epochs for the training. Defaults to 512.
        validate_epoch (int, optional):
            The frequency of validation score measurement. Defaults to 5.
        score_degradation_max (int, optional):
            Maximal number of allowed score degradation. Defaults to 5. Defaults to 5.
    """

    recommender_class: Type[BaseRecommenderWithEarlyStopping]

    def __init__(
        self,
        data: InteractionMatrix,
        val_evaluator: Evaluator,
        logger: Optional[logging.Logger] = None,
        suggest_overwrite: List[Suggestion] = list(),
        fixed_params: Dict[str, Any] = dict(),
        max_epoch: int = 512,
        validate_epoch: int = 5,
        score_degradation_max: int = 5,
        **kwargs: Any,
    ):

        super().__init__(
            data,
            val_evaluator,
            logger=logger,
            suggest_overwrite=suggest_overwrite,
            fixed_params=fixed_params,
        )
        self.max_epoch = max_epoch
        self.validate_epoch = validate_epoch
        self.score_degradation_max = score_degradation_max

    def get_model_arguments(
        self, *args: Any, **kwargs: Any
    ) -> Tuple[Tuple[Any, ...], Dict[str, Any]]:
        return super().get_model_arguments(
            *args,
            max_epoch=self.max_epoch,
            validate_epoch=self.validate_epoch,
            score_degradation_max=self.score_degradation_max,
            **kwargs,
        )


def get_optimizer_class(optimizer_name: str) -> Type[BaseOptimizer]:
    result: Type[BaseOptimizer] = OptimizerMeta.optimizer_name_vs_optimizer_class[
        optimizer_name
    ]
    return result



from typing import Dict

import numpy as np
import pytest
import scipy.sparse as sps


def test_bprFM(test_interaction_data: Dict[str, sps.csr_matrix]) -> None:

    try:
        from irspack.recommenders import BPRFMRecommender
    except:
        pytest.skip("lightfm not found.")
        raise
    X = test_interaction_data["X_small"]
    rec = BPRFMRecommender(X, n_components=min(X.shape), max_epoch=30)
    with pytest.raises(RuntimeError):
        _ = rec.fm.user_embeddings
    rec.learn()
    score = rec.get_user_embedding().dot(rec.get_item_embedding().T)
    assert score.shape == X.shape
    score_2 = rec.get_score_remove_seen_block(0, X.shape[0])
    assert np.all(np.isneginf(score_2[X.nonzero()]))



from pkg_resources import DistributionNotFound, get_distribution

from irspack.recommenders.toppop import TopPopRecommender

try:
    __version__ = get_distribution("irspack").version
except DistributionNotFound:  # pragma: no cover
    # package is not installed
    pass  # pragma: no cover

from irspack.definitions import DenseScoreArray, InteractionMatrix, UserIndexArray
from irspack.evaluator import *
from irspack.optimizers import *
from irspack.recommenders import *
from irspack.split import *



from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import numpy as np
import scipy.sparse as sps

from irspack.definitions import DenseScoreArray, UserIndexArray
from irspack.utils._util_cpp import retrieve_recommend_from_score
from irspack.utils.threading import get_n_threads

if TYPE_CHECKING:
    # We should move this module out of "utils".
    from irspack.recommenders import BaseRecommender


class IDMappedRecommender:
    """A utility class that helps  mapping user/item ids to index, retrieving recommendation score,
    and making a recommendation.

    Args:
        recommender:
            The backend base recommender which transforms user/item ids.
        user_ids:
            user_ids which correspods to the rows of ``recommender.X_train_all``.
        item_ids:
            item_ids which correspods to the columns of ``recommender.X_train_all``.

    Raises:
        ValueError: When recommender and user_ids/item_ids are inconsistent.
        ValueError: When there is a duplicate in user_ids.
        ValueError: When there is a duplicate in item_ids.

    """

    def __init__(
        self, recommender: "BaseRecommender", user_ids: List[Any], item_ids: List[Any]
    ):

        if (recommender.n_users != len(user_ids)) or (
            recommender.n_items != len(item_ids)
        ):
            raise ValueError(
                "The recommender and user/item ids have inconsistent lengths."
            )

        self.recommender = recommender
        self.user_ids = user_ids
        self.item_ids = item_ids
        self.user_id_to_index = {user_id: i for i, user_id in enumerate(user_ids)}
        self.item_id_to_index = {item_id: i for i, item_id in enumerate(item_ids)}

    def _item_id_list_to_index_list(self, ids: Iterable[Any]) -> List[int]:
        return [self.item_id_to_index[id] for id in ids if id in self.item_id_to_index]

    def _user_profile_to_data_col(
        self, profile: Union[List[Any], Dict[Any, float]]
    ) -> Tuple[List[float], List[int]]:
        data: List[float]
        cols: List[int]
        # data: np.ndarray
        if isinstance(profile, list):
            cols = self._item_id_list_to_index_list(profile)
            data = [1.0] * len(cols)
        else:
            data = []
            cols = []
            for id, score in profile.items():
                if id in self.item_id_to_index:
                    data.append(score)
                    cols.append(self.item_id_to_index[id])
        return data, cols

    def _list_of_user_profile_to_matrix(
        self, users_info: Sequence[Union[List[Any], Dict[Any, float]]]
    ) -> sps.csr_matrix:
        data: List[float] = []
        indptr: List[int] = [0]
        col: List[int] = []
        indptr_current = 0
        for user_info in users_info:
            data_u, col_u = self._user_profile_to_data_col(user_info)
            data.extend(data_u)
            col.extend(col_u)
            indptr_current += len(col_u)
            indptr.append(indptr_current)
        result = sps.csr_matrix(
            (data, col, indptr), shape=(len(users_info), len(self.item_ids))
        )
        return result

    def get_recommendation_for_known_user_id(
        self,
        user_id: Any,
        cutoff: int = 20,
        allowed_item_ids: Optional[List[Any]] = None,
        forbidden_item_ids: Optional[List[Any]] = None,
    ) -> List[Tuple[Any, float]]:
        """Retrieve recommendation result for a known user.
        Args:
            user_id:
                The target user ID.
            cutoff:
                Maximal number of recommendations allowed.
            allowed_item_ids:
                If not ``None``, recommend the items within this list.
                If ``None``, all known item ids can be recommended (except for those in ``item_ids`` argument).
                Defaults to ``None``.
            forbidden_item_ids:
                If not ``None``, never recommend the items within the list. Defaults to None.

        Raises:
            RuntimeError: When user_id is not in ``self.user_ids``.

        Returns:
            A List of tuples consisting of ``(item_id, score)``.
        """
        if user_id not in self.user_ids:
            raise RuntimeError(f"User with user_id {user_id} not found.")
        user_index: UserIndexArray = np.asarray(
            [self.user_id_to_index[user_id]], dtype=np.int64
        )

        score = self.recommender.get_score_remove_seen(user_index)[0, :]
        return self._score_to_recommended_items(
            score,
            cutoff=cutoff,
            allowed_item_ids=allowed_item_ids,
            forbidden_item_ids=forbidden_item_ids,
        )

    def get_recommendation_for_known_user_batch(
        self,
        user_ids: List[Any],
        cutoff: int = 20,
        allowed_item_ids: Optional[List[List[Any]]] = None,
        forbidden_item_ids: Optional[List[List[Any]]] = None,
        n_threads: Optional[int] = None,
    ) -> List[List[Tuple[Any, float]]]:
        """Retrieve recommendation result for a list of known users.

        Args:
            user_ids:
                A list of user ids.
            cutoff:
                Maximal number of recommendations allowed.
            allowed_item_ids:
                If not ``None``, defines "a list of list of recommendable item IDs"
                and ``len(allowed_item_ids)`` must be equal to ``len(item_ids)``.
                Defaults to ``None``.
            forbidden_item_ids:
                If not ``None``, defines "a list of list of forbidden item IDs"
                and ``len(allowed_item_ids)`` must be equal to ``len(item_ids)``
                Defaults to ``None``.

        Returns:
            A list of list of tuples consisting of ``(item_id, score)``.
            Each internal list corresponds to the recommender's recommendation output.
        """
        user_indexes: UserIndexArray = np.asarray(
            [self.user_id_to_index[user_id] for user_id in user_ids], dtype=np.int64
        )

        score = self.recommender.get_score_remove_seen(user_indexes)
        return self._score_to_recommended_items_batch(
            score,
            cutoff=cutoff,
            allowed_item_ids=allowed_item_ids,
            forbidden_item_ids=forbidden_item_ids,
            n_threads=get_n_threads(n_threads=n_threads),
        )

    def get_recommendation_for_new_user(
        self,
        user_profile: Union[List[Any], Dict[Any, float]],
        cutoff: int = 20,
        allowed_item_ids: Optional[List[Any]] = None,
        forbidden_item_ids: Optional[List[Any]] = None,
    ) -> List[Tuple[Any, float]]:
        """Retrieve recommendation result for a previously unseen user using item ids with which he or she interacted.

        Args:
            user_profile:
                User's profile given either as a list of item ids the user had a cotact or a item id-rating dict.
                Previously unseen item ID will be ignored.
            cutoff:
                Maximal number of recommendations allowed.
            allowed_item_ids:
                If not ``None``, recommend the items within this list.
                If ``None``, all known item ids can be recommended (except for those in ``item_ids`` argument).
                Defaults to ``None``.
            forbidden_item_ids:
                If not ``None``, never recommend the items within the list. Defaults to None.

        Returns:
            A List of tuples consisting of ``(item_id, score)``.
        """
        data, cols = self._user_profile_to_data_col(user_profile)
        X_input = sps.csr_matrix(
            (data, cols, [0, len(cols)]), shape=(1, len(self.item_ids))
        )
        score = self.recommender.get_score_cold_user_remove_seen(X_input)[0]
        return self._score_to_recommended_items(
            score,
            cutoff,
            allowed_item_ids=allowed_item_ids,
            forbidden_item_ids=forbidden_item_ids,
        )

    def get_recommendation_for_new_user_batch(
        self,
        user_profiles: Sequence[Union[List[Any], Dict[Any, float]]],
        cutoff: int = 20,
        allowed_item_ids: Optional[List[List[Any]]] = None,
        forbidden_item_ids: Optional[List[List[Any]]] = None,
        n_threads: Optional[int] = None,
    ) -> List[List[Tuple[Any, float]]]:
        """Retrieve recommendation result for a previously unseen users using item ids with which they have interacted.

        Args:
            user_profiles:
                A list of user profiles.
                Each profile should be either the item ids the user had a cotact, or item-rating dict.
                Previously unseen item IDs will be ignored.
            cutoff:
                Maximal number of recommendations allowed.
            allowed_item_ids:
                If not ``None``, defines "a list of list of recommendable item IDs"
                and ``len(allowed_item_ids)`` must be equal to ``len(item_ids)``.
                Defaults to ``None``.
            forbidden_item_ids:
                If not ``None``, defines "a list of list of forbidden item IDs"
                and ``len(allowed_item_ids)`` must be equal to ``len(item_ids)``
                Defaults to ``None``.

        Returns:
            A list of list of tuples consisting of ``(item_id, score)``.
            Each internal list corresponds to the recommender's recommendation output.
        """
        X_input = self._list_of_user_profile_to_matrix(user_profiles)
        score = self.recommender.get_score_cold_user_remove_seen(X_input)
        return self._score_to_recommended_items_batch(
            score,
            cutoff,
            allowed_item_ids=allowed_item_ids,
            forbidden_item_ids=forbidden_item_ids,
            n_threads=get_n_threads(n_threads=n_threads),
        )

    def _score_to_recommended_items(
        self,
        score: DenseScoreArray,
        cutoff: int,
        allowed_item_ids: Optional[List[Any]] = None,
        forbidden_item_ids: Optional[List[Any]] = None,
    ) -> List[Tuple[Any, float]]:
        if allowed_item_ids is not None:
            allowed_item_indices = np.asarray(
                self._item_id_list_to_index_list(allowed_item_ids), dtype=np.int64
            )
            high_score_inds = allowed_item_indices[
                score[allowed_item_indices].argsort()[::-1]
            ]
        else:
            high_score_inds = score.argsort()[::-1]
        recommendations: List[Tuple[Any, float]] = []
        for i in high_score_inds:
            i_int = int(i)
            score_this = score[i_int]
            item_id = self.item_ids[i_int]
            if np.isinf(score_this):
                continue
            if forbidden_item_ids is not None:
                if item_id in forbidden_item_ids:
                    continue
            recommendations.append((item_id, float(score_this)))
            if len(recommendations) >= cutoff:
                break
        return recommendations

    def _score_to_recommended_items_batch(
        self,
        score: DenseScoreArray,
        cutoff: int,
        allowed_item_ids: Optional[List[List[Any]]] = None,
        forbidden_item_ids: Optional[List[List[Any]]] = None,
        n_threads: int = 1,
    ) -> List[List[Tuple[Any, float]]]:
        if forbidden_item_ids is not None:
            assert len(forbidden_item_ids) == score.shape[0]
        if allowed_item_ids is not None:
            assert len(allowed_item_ids) == score.shape[0]

        allowed_item_indices: List[List[int]] = []
        if allowed_item_ids is not None:
            allowed_item_indices = [
                self._item_id_list_to_index_list(_) for _ in allowed_item_ids
            ]
        if forbidden_item_ids is not None:
            for u, forbidden_ids_per_user in enumerate(forbidden_item_ids):
                score[
                    u, self._item_id_list_to_index_list(forbidden_ids_per_user)
                ] = -np.inf

        raw_result = retrieve_recommend_from_score(
            score,
            allowed_item_indices,
            cutoff,
            n_threads=n_threads,
        )
        return [
            [
                (self.item_ids[item_index], score)
                for item_index, score in user_wise_raw_result
            ]
            for user_wise_raw_result in raw_result
        ]



from typing import Dict

import numpy as np
import pytest
import scipy.sparse as sps
from sklearn.linear_model import ElasticNet

from irspack.recommenders import SLIMRecommender


def test_slim_positive(test_interaction_data: Dict[str, sps.csr_matrix]) -> None:
    alpha = 0.1
    l1_ratio = 0.5
    X = test_interaction_data["X_small"]
    rec = SLIMRecommender(
        X,
        alpha=alpha,
        l1_ratio=l1_ratio,
        positive_only=True,
        n_iter=100,
        n_threads=8,
        tol=0,
    )
    rec.learn()

    enet = ElasticNet(
        alpha=alpha,
        l1_ratio=l1_ratio,
        fit_intercept=False,
        positive=True,
        max_iter=100,
        tol=1e-8,
    )
    for iind in range(rec.W.shape[1]):
        m = rec.W[:, iind].toarray().ravel()
        Xcp = X.toarray()
        y = X[:, iind].toarray().ravel()
        Xcp[:, iind] = 0.0
        enet.fit(Xcp, y)
        np.testing.assert_allclose(enet.coef_, m, rtol=1e-2)


def test_slim_allow_negative(test_interaction_data: Dict[str, sps.csr_matrix]) -> None:
    ITER = 100
    alpha = 0.1
    l1_ratio = 0.5
    X = test_interaction_data["X_small"]
    rec = SLIMRecommender(
        X,
        alpha=alpha,
        l1_ratio=l1_ratio,
        positive_only=False,
        n_iter=ITER,
        n_threads=8,
        tol=0,
    )
    rec.learn()

    enet = ElasticNet(
        alpha=alpha, l1_ratio=l1_ratio, fit_intercept=False, max_iter=ITER, tol=1e-8
    )
    for iind in range(rec.W.shape[1]):
        m = rec.W[:, iind].toarray().ravel()
        Xcp = X.toarray()
        y = X[:, iind].toarray().ravel()
        Xcp[:, iind] = 0.0
        enet.fit(Xcp, y)
        np.testing.assert_allclose(enet.coef_, m, rtol=1e-2)


def test_slim_topk(test_interaction_data: Dict[str, sps.csr_matrix]) -> None:
    alpha = 0.0001
    l1_ratio = 0.0
    X = test_interaction_data["X_small"]
    rec = SLIMRecommender(
        X,
        alpha=alpha,
        l1_ratio=l1_ratio,
        positive_only=True,
        n_iter=100,
        n_threads=1,
        tol=0,
    )
    rec.learn()
    W_non_restricted = rec.W.toarray()

    rec_restricted = SLIMRecommender(
        X,
        alpha=alpha,
        l1_ratio=l1_ratio,
        positive_only=True,
        n_iter=100,
        n_threads=1,
        tol=0,
        top_k=1,
    )
    rec_restricted.learn()
    W_restricted = rec_restricted.W.toarray()
    for i in range(rec.n_items):
        gt = W_non_restricted[:, i]
        target = W_restricted[:, i]
        assert np.sum(target > 0) <= 1
        target.max() == pytest.approx(gt.max())



import os
import time
from typing import Dict, List

import numpy as np
import pandas as pd
import pytest
from scipy import sparse as sps

from irspack import (
    BaseOptimizer,
    BaseRecommender,
    DenseScoreArray,
    InteractionMatrix,
    UserIndexArray,
    autopilot,
)
from irspack.evaluator.evaluator import Evaluator
from irspack.parameter_tuning import Suggestion, UniformSuggestion

SKIP_TEST = os.environ.get("IRSPACK_TESTING", "false") != "true"

TIMESCALE = 1.5

X_small = sps.csr_matrix(
    (np.random.RandomState(42).rand(10, 1024) > 0.5).astype(np.float64)
)
X_answer = sps.csr_matrix(
    (np.random.RandomState(43).rand(*X_small.shape) > 0.5).astype(np.float64)
)


class AutopilotMockRecommender(BaseRecommender):
    def __init__(self, X: InteractionMatrix, wait_time: float = 1.0):
        super().__init__(X)
        self.wait_time = wait_time
        self.leaked = X_answer.toarray()

    def get_score(self, user_indices: UserIndexArray) -> DenseScoreArray:
        return self.leaked[user_indices]

    def _learn(self) -> None:
        time.sleep(self.wait_time)


class AutoPilotMockOptimizer(BaseOptimizer):
    recommender_class = AutopilotMockRecommender
    default_tune_range: List[Suggestion] = [UniformSuggestion("wait_time", 0.1, 4.0)]

    @classmethod
    def tune_range_given_memory_budget(
        cls, X: InteractionMatrix, memory_in_mb: int
    ) -> List[Suggestion]:
        return []


@pytest.mark.skipif(SKIP_TEST, reason="May not work on non-Linux.")
def test_autopilot() -> None:
    evaluator = Evaluator(X_answer, 0)
    recommender_class, best_param, trial_df = autopilot(
        X_small,
        evaluator,
        memory_budget=1,
        n_trials=10,
        algorithms=["AutoPilotMock"],
        timeout_singlestep=2,
    )
    assert best_param["wait_time"] < 2.0
    assert trial_df.shape[0] == 10
    wait_times = trial_df["AutoPilotMockOptimizer.wait_time"]
    assert np.all(trial_df.iloc[(wait_times.values > 2.0)]["ndcg@10"].isna())
    assert recommender_class is AutopilotMockRecommender


@pytest.mark.skipif(SKIP_TEST, reason="May not work on non-Linux.")
def test_autopilot_timeout() -> None:
    evaluator = Evaluator(X_answer, 0)
    wait = 20
    with pytest.raises(RuntimeError):
        # no available algorithm
        autopilot(
            X_small,
            evaluator,
            memory_budget=1,
            n_trials=100,
            algorithms=["DenseSLIM"],
            timeout_overall=5,
            timeout_singlestep=1,
        )

    start = time.time()

    wait_times_given_by_callback: Dict[int, float] = {}

    def callback(trial_number: int, history_df: pd.DataFrame) -> None:
        wait_times_given_by_callback[trial_number] = history_df[
            "AutoPilotMockOptimizer.wait_time"
        ].iloc[-1]

    recommender_class, best_param, trial_df = autopilot(
        X_small,
        evaluator,
        memory_budget=1,
        n_trials=100,
        algorithms=["AutoPilotMock", "DenseSLIM"],
        timeout_overall=wait,
        timeout_singlestep=1,
        callback=callback,
    )
    end = time.time()
    assert wait < (end - start) + 1
    wait_times = trial_df["AutoPilotMockOptimizer.wait_time"]
    assert np.all(trial_df.iloc[(wait_times.values > 1.0)]["ndcg@10"].isna())
    # dense slim should be skipped
    assert len({name for name in trial_df["optimizer_name"] if not pd.isna(name)}) == 1
    assert recommender_class is AutopilotMockRecommender
    recommender_class(X_small, **best_param).learn()
    for index, row in trial_df.iterrows():
        target_value = row["AutoPilotMockOptimizer.wait_time"]
        test_value = wait_times_given_by_callback[index]
        if pd.isna(test_value):
            assert pd.isna(target_value)
        else:
            assert target_value == test_value



import warnings
from typing import List, Type

from irspack.definitions import InteractionMatrix

from ..optimizers.base_optimizer import (
    BaseOptimizer,
    BaseOptimizerWithEarlyStopping,
    LowMemoryError,
)
from ..parameter_tuning import (
    CategoricalSuggestion,
    IntegerSuggestion,
    LogUniformSuggestion,
    Suggestion,
    UniformSuggestion,
)
from ..recommenders import (
    AsymmetricCosineKNNRecommender,
    AsymmetricCosineUserKNNRecommender,
    CosineKNNRecommender,
    CosineUserKNNRecommender,
    DenseSLIMRecommender,
    IALSRecommender,
    JaccardKNNRecommender,
    NMFRecommender,
    P3alphaRecommender,
    RP3betaRecommender,
    SLIMRecommender,
    TopPopRecommender,
    TruncatedSVDRecommender,
    TverskyIndexKNNRecommender,
)

default_tune_range_knn = [
    IntegerSuggestion("top_k", 4, 1000),
    UniformSuggestion("shrinkage", 0, 1000),
]

default_tune_range_knn_with_weighting = [
    IntegerSuggestion("top_k", 4, 1000),
    UniformSuggestion("shrinkage", 0, 1000),
    CategoricalSuggestion("feature_weighting", ["NONE", "TF_IDF", "BM_25"]),
]


def _get_maximal_n_components_for_budget(
    X: InteractionMatrix, budget: int, n_component_max_default: int
) -> int:
    return min(int((budget * 1e6) / (8 * (sum(X.shape) + 1))), n_component_max_default)


class TopPopOptimizer(BaseOptimizer):
    default_tune_range: List[Suggestion] = []
    recommender_class = TopPopRecommender

    @classmethod
    def tune_range_given_memory_budget(
        cls, X: InteractionMatrix, memory_budget: int
    ) -> List[Suggestion]:
        return []


class IALSOptimizer(BaseOptimizerWithEarlyStopping):
    default_tune_range = [
        IntegerSuggestion("n_components", 4, 300),
        LogUniformSuggestion("alpha", 1, 100),
        LogUniformSuggestion("reg", 1e-10, 1e-2),
    ]
    recommender_class = IALSRecommender

    @classmethod
    def tune_range_given_memory_budget(
        cls, X: InteractionMatrix, memory_budget: int
    ) -> List[Suggestion]:
        n_components = _get_maximal_n_components_for_budget(X, memory_budget, 300)
        return [
            IntegerSuggestion("n_components", 4, n_components),
        ]


class DenseSLIMOptimizer(BaseOptimizer):
    default_tune_range: List[Suggestion] = [LogUniformSuggestion("reg", 1, 1e4)]
    recommender_class = DenseSLIMRecommender

    @classmethod
    def tune_range_given_memory_budget(
        cls, X: InteractionMatrix, memory_budget: int
    ) -> List[Suggestion]:
        n_items: int = X.shape[1]
        if (1e6 * memory_budget) < (4 * 2 * n_items ** 2):
            raise LowMemoryError(
                f"Memory budget {memory_budget} too small for DenseSLIM to work."
            )
        return []


class TruncatedSVDOptimizer(BaseOptimizer):
    default_tune_range = [IntegerSuggestion("n_components", 4, 512)]
    recommender_class = TruncatedSVDRecommender

    @classmethod
    def tune_range_given_memory_budget(
        cls, X: InteractionMatrix, memory_budget: int
    ) -> List[Suggestion]:
        n_components = _get_maximal_n_components_for_budget(X, memory_budget, 512)
        return [
            IntegerSuggestion("n_components", 4, n_components),
        ]


class NMFOptimizer(BaseOptimizer):
    default_tune_range = [
        IntegerSuggestion("n_components", 4, 512),
        LogUniformSuggestion("alpha", 1e-10, 1e-1),
        UniformSuggestion("l1_ratio", 0, 1),
    ]
    recommender_class = NMFRecommender

    @classmethod
    def tune_range_given_memory_budget(
        cls, X: InteractionMatrix, memory_budget: int
    ) -> List[Suggestion]:
        n_components = _get_maximal_n_components_for_budget(X, memory_budget, 512)
        return [
            IntegerSuggestion("top_k", 4, n_components),
        ]


class SimilarityBasedOptimizerBase(BaseOptimizer):
    @classmethod
    def tune_range_given_memory_budget(
        cls, X: InteractionMatrix, memory_budget: int
    ) -> List[Suggestion]:
        top_k_max = min(int(1e6 * memory_budget / 4 // (X.shape[1] + 1)), 1024)
        if top_k_max <= 4:
            raise LowMemoryError(
                f"Memory budget {memory_budget} too small for {cls.__name__} to work."
            )

        return [
            IntegerSuggestion("top_k", 4, top_k_max),
        ]


class SLIMOptimizer(SimilarityBasedOptimizerBase):
    default_tune_range = [
        LogUniformSuggestion("alpha", 1e-5, 1),
        UniformSuggestion("l1_ratio", 0, 1),
    ]
    recommender_class = SLIMRecommender


class P3alphaOptimizer(SimilarityBasedOptimizerBase):
    default_tune_range = [
        IntegerSuggestion("top_k", low=10, high=1000),
        CategoricalSuggestion("normalize_weight", [True, False]),
    ]
    recommender_class = P3alphaRecommender


class RP3betaOptimizer(SimilarityBasedOptimizerBase):
    default_tune_range = [
        IntegerSuggestion("top_k", 2, 1000),
        LogUniformSuggestion("beta", 1e-5, 5e-1),
        CategoricalSuggestion("normalize_weight", [True, False]),
    ]
    recommender_class = RP3betaRecommender


class CosineKNNOptimizer(SimilarityBasedOptimizerBase):
    default_tune_range = default_tune_range_knn_with_weighting.copy() + [
        CategoricalSuggestion("normalize", [False, True])
    ]

    recommender_class = CosineKNNRecommender


class AsymmetricCosineKNNOptimizer(SimilarityBasedOptimizerBase):
    default_tune_range = default_tune_range_knn_with_weighting + [
        UniformSuggestion("alpha", 0, 1)
    ]

    recommender_class = AsymmetricCosineKNNRecommender


class JaccardKNNOptimizer(SimilarityBasedOptimizerBase):

    default_tune_range = default_tune_range_knn.copy()
    recommender_class = JaccardKNNRecommender


class TverskyIndexKNNOptimizer(SimilarityBasedOptimizerBase):
    default_tune_range = default_tune_range_knn.copy() + [
        UniformSuggestion("alpha", 0, 2),
        UniformSuggestion("beta", 0, 2),
    ]

    recommender_class = TverskyIndexKNNRecommender


class UserSimilarityBasedOptimizerBase(BaseOptimizer):
    @classmethod
    def tune_range_given_memory_budget(
        cls, X: InteractionMatrix, memory_budget: int
    ) -> List[Suggestion]:
        top_k_max = min(int(1e6 * memory_budget / 4 // (X.shape[0] + 1)), 1024)
        return [
            IntegerSuggestion("top_k", 4, top_k_max),
        ]


class CosineUserKNNOptimizer(UserSimilarityBasedOptimizerBase):
    default_tune_range = default_tune_range_knn_with_weighting.copy() + [
        CategoricalSuggestion("normalize", [False, True])
    ]

    recommender_class = CosineUserKNNRecommender


class AsymmetricCosineUserKNNOptimizer(UserSimilarityBasedOptimizerBase):
    default_tune_range = default_tune_range_knn_with_weighting + [
        UniformSuggestion("alpha", 0, 1)
    ]

    recommender_class = AsymmetricCosineUserKNNRecommender


try:
    from ..recommenders.bpr import BPRFMRecommender

    class BPRFMOptimizer(BaseOptimizerWithEarlyStopping):
        default_tune_range = [
            IntegerSuggestion("n_components", 4, 256),
            LogUniformSuggestion("item_alpha", 1e-9, 1e-2),
            LogUniformSuggestion("user_alpha", 1e-9, 1e-2),
            CategoricalSuggestion("loss", ["bpr", "warp"]),
        ]
        recommender_class = BPRFMRecommender

        @classmethod
        def tune_range_given_memory_budget(
            cls, X: InteractionMatrix, memory_budget: int
        ) -> List[Suggestion]:
            # memory usage will be roughly 4 (float) * (n_users + n_items) * k
            n_components = _get_maximal_n_components_for_budget(X, memory_budget, 300)
            return [
                IntegerSuggestion("n_components", 4, n_components),
            ]


except:  # pragma: no cover
    pass  # pragma: no cover


try:
    from ..recommenders.multvae import MultVAERecommender

    class MultVAEOptimizer(BaseOptimizerWithEarlyStopping):
        default_tune_range = [
            CategoricalSuggestion("dim_z", [32, 64, 128, 256]),
            CategoricalSuggestion("enc_hidden_dims", [128, 256, 512]),
            CategoricalSuggestion("kl_anneal_goal", [0.1, 0.2, 0.4]),
        ]
        recommender_class = MultVAERecommender

        @classmethod
        def tune_range_given_memory_budget(
            cls, X: InteractionMatrix, memory_budget: int
        ) -> List[Suggestion]:
            if memory_budget * 1e6 > (X.shape[1] * 2048 * 8):
                raise LowMemoryError(
                    f"Memory budget {memory_budget} too small for MultVAE to work."
                )

            return []


except:  # pragma: no cover
    warnings.warn("MultVAEOptimizer is not available.")  # pragma: no cover



from typing import Optional

from irspack.definitions import InteractionMatrix
from irspack.recommenders.base import BaseSimilarityRecommender
from irspack.utils import get_n_threads
from irspack.utils._util_cpp import (
    slim_weight_allow_negative,
    slim_weight_positive_only,
)


class SLIMRecommender(BaseSimilarityRecommender):
    r"""`SLIM <https://dl.acm.org/doi/10.1109/ICDM.2011.134>`_ with ElasticNet-type loss function:

    .. math ::

        \mathrm{loss} = \frac{1}{2} ||X - XB|| ^2 _F + \frac{\alpha (1 - l_1)  U}{2} ||B|| ^2 _FF + \alpha l_1  U |B|

    The implementation relies on a simple (parallelized) cyclic-coordinate descent method.

    Args:
        X_train_all:
            Input interaction matrix.
        alpha:
            Determines the strength of L1/L2 regularization (see above). Defaults to 0.05.
        l1_ratio:
            Determines the strength of L1 regularization relative to alpha. Defaults to 0.01.
        positive_only:
            Whether we constrain the weight matrix to be non-negative. Defaults to True.
        n_iter:
            The number of coordinate-descent iterations. Defaults to 100.
        tol:
            Tolerance parameter for cd iteration, i.e., if the maximal parameter change
            of the coordinate-descent single iteration is smaller than this value,
            the iteration will terminate. Defaults to 1e-4.
        top_k:
            Specifies the maximal number of allowed non-zero coefficients per item. Defaults to None.
        n_threads:
            Specifies the number of threads to use for the computation.
            If ``None``, the environment variable ``"IRSPACK_NUM_THREADS_DEFAULT"`` will be looked up,
            and if there is no such an environment variable, it will be set to ``os.cpu_count()``. Defaults to None.
    """

    def __init__(
        self,
        X_train_all: InteractionMatrix,
        alpha: float = 0.05,
        l1_ratio: float = 0.01,
        positive_only: bool = True,
        n_iter: int = 100,
        tol: float = 1e-4,
        top_k: Optional[int] = None,
        n_threads: Optional[int] = None,
    ):
        super().__init__(X_train_all)
        self.alpha = alpha
        self.l1_ratio = l1_ratio
        self.positive_only = positive_only
        self.n_threads = get_n_threads(n_threads)
        self.n_iter = n_iter
        self.tol = tol
        self.top_k = top_k

    def _learn(self) -> None:
        l2_coeff = self.n_users * self.alpha * (1 - self.l1_ratio)
        l1_coeff = self.n_users * self.alpha * self.l1_ratio

        if self.positive_only:
            self.W_ = slim_weight_positive_only(
                self.X_train_all,
                n_threads=self.n_threads,
                n_iter=self.n_iter,
                l2_coeff=l2_coeff,
                l1_coeff=l1_coeff,
                tol=self.tol,
                top_k=-1 if self.top_k is None else self.top_k,
            )
        else:
            self.W_ = slim_weight_allow_negative(
                self.X_train_all,
                n_threads=self.n_threads,
                n_iter=self.n_iter,
                l2_coeff=l2_coeff,
                l1_coeff=l1_coeff,
                tol=self.tol,
                top_k=-1 if self.top_k is None else self.top_k,
            )



import numpy as np
import pytest
import scipy.sparse as sps

from irspack.definitions import InteractionMatrix, ProfileMatrix
from irspack.evaluator import Evaluator
from irspack.split import rowwise_train_test_split
from irspack.user_cold_start.evaluator import UserColdStartEvaluator
from irspack.user_cold_start.optimizers import LinearMethodOptimizer
from irspack.user_cold_start.recommenders import LinearMethodRecommender

RNS = np.random.RandomState(0)

profile = sps.csr_matrix(RNS.rand(3, 20) >= 0.7).astype(np.float64)
X_cf = sps.csr_matrix(RNS.rand(3, 10) >= 0.7).astype(np.float64)

profile = sps.vstack([profile for _ in range(100)])  # so many duplicates!
X_cf = sps.vstack([X_cf for _ in range(100)])


def test_lineamethod() -> None:
    config = LinearMethodOptimizer.split_and_optimize(
        X_cf, profile, evaluator_config=dict(cutoff=10), random_seed=0
    )
    rec = LinearMethodRecommender(X_cf, profile, **config).learn()
    evaluator = UserColdStartEvaluator(X_cf, profile, cutoff=5)
    scores = evaluator.get_score(rec)
    print(scores, config)
    assert scores["ndcg"] >= 0.8



import os

import pandas as pd

from .base import BaseMovieLenstDataLoader


class MovieLens1MDataManager(BaseMovieLenstDataLoader):
    DOWNLOAD_URL = "http://files.grouplens.org/datasets/movielens/ml-1m.zip"
    DEFAULT_PATH = os.path.expanduser("~/.ml-1m.zip")
    INTERACTION_PATH = "ml-1m/ratings.dat"
    ITEM_INFO_PATH = "ml-1m/movies.dat"
    USER_INFO_PATH = "ml-1m/users.dat"

    def read_interaction(self) -> pd.DataFrame:
        with self._read_as_istream(self.INTERACTION_PATH) as ifs:
            # This is a hack.
            # The true separator is "::", but this will force pandas
            # to use python engine, which is much slower.
            # instead we regard the separator to be ':' and imagine there is an empty (NaN).
            # values between "::"
            df = pd.read_csv(
                ifs,
                sep=":",
                header=None,
            )[[0, 2, 4, 6]].copy()

            df.columns = ["userId", "movieId", "rating", "timestamp"]
            df["timestamp"] = pd.to_datetime(df.timestamp, unit="s")
            return df

    def read_item_info(self) -> pd.DataFrame:
        with self._read_as_istream(self.ITEM_INFO_PATH) as ifs:
            data = pd.read_csv(
                ifs,
                sep="::",
                header=None,
                encoding="latin-1",
                names=["movieId", "title", "genres"],
                engine="python",
            )
            release_year = pd.to_numeric(
                data.title.str.extract(r"^.*\((?P<release_year>\d+)\)\s*$").release_year
            )
            data["release_year"] = release_year
            return data.set_index("movieId")

    def read_user_info(self) -> pd.DataFrame:
        with self._read_as_istream(self.USER_INFO_PATH) as ifs:
            return pd.read_csv(
                ifs,
                sep="::",
                header=None,
                names=["userId", "gender", "age", "occupation", "zipcode"],
                engine="python",
            ).set_index("userId")



import numpy as np
import pytest
import scipy.sparse as sps

from irspack.recommenders.user_knn import (
    AsymmetricCosineUserKNNRecommender,
    CosineUserKNNRecommender,
)

X_small = sps.csr_matrix(
    np.asfarray([[1, 1, 2, 3, 4], [0, 1, 0, 1, 0], [0, 0, 1, 0, 0], [0, 0, 0, 0, 0]])
)
X_many = np.random.rand(888, 512)
X_many[X_many <= 0.9] = 0
X_many[X_many > 0.9] = 1
X_many = sps.csr_matrix(X_many)
X_many.sort_indices()

X_many_dense = sps.csr_matrix(np.random.rand(133, 245))


@pytest.mark.parametrize(
    "X, normalize", [(X_many, True), (X_small, False), (X_many_dense, True)]
)
def test_cosine(X: sps.csr_matrix, normalize: bool) -> None:
    rec = CosineUserKNNRecommender(
        X, shrinkage=0, n_threads=5, top_k=X.shape[0], normalize=normalize
    )
    with pytest.raises(RuntimeError):
        U = rec.U
    rec.learn()
    sim = rec.U.toarray()
    manual = X.toarray()  # U x I
    norm = (manual ** 2).sum(axis=1) ** 0.5
    manual = manual.dot(manual.T)
    if normalize:
        denom = norm[:, None] * norm[None, :] + 1e-6
        manual /= denom
    np.fill_diagonal(manual, 0)
    np.testing.assert_allclose(
        sim,
        manual,
    )


@pytest.mark.parametrize(
    "X, alpha, shrinkage",
    [(X_many, 0.5, 0.0), (X_small, 0.7, 1.0), (X_many_dense, 0.01, 3)],
)
def test_asymmetric_cosine(X: sps.csr_matrix, alpha: float, shrinkage: float) -> None:
    rec = AsymmetricCosineUserKNNRecommender(
        X, shrinkage=shrinkage, alpha=alpha, n_threads=1, top_k=X.shape[0]
    )
    rec.learn()
    sim = rec.U.toarray()

    manual = X.toarray()
    norm = (manual ** 2).sum(axis=1)
    norm_alpha = np.power(norm, alpha)
    norm_1malpha = np.power(norm, 1 - alpha)
    manual_sim = manual.dot(manual.T)
    denom = norm_alpha[:, None] * norm_1malpha[None, :] + 1e-6 + shrinkage
    manual_sim /= denom
    np.fill_diagonal(manual_sim, 0)
    np.testing.assert_allclose(
        sim,
        manual_sim,
    )


@pytest.mark.parametrize("X", [X_many, X_small])
def test_topk(X: sps.csr_matrix) -> None:
    rec = AsymmetricCosineUserKNNRecommender(X, shrinkage=0, top_k=30, n_threads=5)
    rec.learn()
    sim = rec.U.toarray()
    assert np.all((sim > 0).sum(axis=1) <= 30)



from typing import Optional

import numpy as np
from scipy import sparse as sps

from irspack.definitions import DenseScoreArray
from irspack.recommenders._knn import CosineSimilarityComputer
from irspack.utils import get_n_threads, sparse_mm_threaded

from .base import BaseUserColdStartRecommender, InteractionMatrix, ProfileMatrix


class UserCBCosineKNNRecommender(BaseUserColdStartRecommender):
    def __init__(
        self,
        X_interaction: InteractionMatrix,
        X_profile: ProfileMatrix,
        top_k: int = 100,
        n_threads: Optional[int] = None,
        shrink: float = 1e-1,
    ):

        self.top_k = top_k
        self.shrink = shrink
        self.n_threads = get_n_threads(n_threads)
        super().__init__(X_interaction, X_profile)
        self.X_interaction_csc = X_interaction.astype(np.float64).tocsc()
        self.sim_computer: Optional[CosineSimilarityComputer] = None

    def _learn(self) -> None:
        self.sim_computer = CosineSimilarityComputer(
            self.X_profile, self.shrink, normalize=True, n_threads=self.n_threads
        )

    def get_score(self, profile: ProfileMatrix) -> DenseScoreArray:
        if self.sim_computer is None:
            raise RuntimeError("'get_score' called before learn.")
        if not sps.issparse(profile):
            profile = sps.csr_matrix(profile)
        similarity = self.sim_computer.compute_similarity(profile, self.top_k)
        score = sparse_mm_threaded(similarity, self.X_interaction_csc, self.n_threads)
        return score.astype(np.float64)



import os
import pickle
from contextlib import redirect_stderr, redirect_stdout
from logging import getLogger
from time import sleep
from typing import IO, Any, Dict

import numpy as np
import pytest
import scipy.sparse as sps

from irspack import (
    BaseOptimizerWithEarlyStopping,
    BaseRecommender,
    BaseRecommenderWithEarlyStopping,
    DenseScoreArray,
    Evaluator,
    InteractionMatrix,
    TargetMetric,
    UserIndexArray,
)
from irspack.parameter_tuning import UniformSuggestion
from irspack.recommenders.base_earlystop import TrainerBase

X_small = sps.csr_matrix(
    (np.random.RandomState(42).rand(100, 32) > 0.5).astype(np.float64)
)
X_large = sps.csr_matrix(
    (np.random.RandomState(42).rand(50, 200) > 0.5).astype(np.float64)
)


class MockTrainer(TrainerBase):
    def __init__(self) -> None:
        self.epoch: int = 0

    def load_state(self, ifs: IO) -> None:
        self.epoch = pickle.load(ifs)["epoch"]

    def save_state(self, ofs: IO) -> None:
        pickle.dump(dict(epoch=self.epoch), ofs)

    def run_epoch(self) -> None:
        self.epoch += 1


class MockRecommender(BaseRecommenderWithEarlyStopping):
    trainer_class = MockTrainer

    def __init__(
        self,
        X: InteractionMatrix,
        target_epoch: int = 20,
        target_score: float = 0.0,
        **kwargs: Any,
    ):
        super().__init__(X, **kwargs)
        self.target_epoch = target_epoch
        self.target_score = target_score
        self.rns = np.random.RandomState(42)

    def _create_trainer(self) -> MockTrainer:
        sleep(0.01)
        return MockTrainer()

    def get_score(self, user_indices: UserIndexArray) -> DenseScoreArray:
        score = np.zeros(*self.X_train_all.shape)
        return score

    def _current_score(self) -> float:
        assert isinstance(self.trainer, MockTrainer)
        coeff: float
        if self.trainer.epoch > self.target_epoch:
            coeff = 1 - (self.trainer.epoch - self.target_epoch) / self.target_epoch
        else:
            coeff = self.trainer.epoch / self.target_epoch
        return self.target_score * coeff


class MockEvaluator(Evaluator):
    def __init__(self, X: sps.csr_matrix) -> None:
        super().__init__(X, offset=0)
        self.target_metric = TargetMetric.ndcg
        self.cutoff = 30

    def get_score(self, model: BaseRecommender) -> Dict[str, float]:
        assert isinstance(model, MockRecommender)
        return {self.target_metric.name: model._current_score()}


class MockOptimizer(BaseOptimizerWithEarlyStopping):
    recommender_class = MockRecommender
    default_tune_range = [UniformSuggestion("target_score", 0, 1)]


@pytest.mark.parametrize("X, target_epoch", [(X_small, 20)])
def test_optimizer_by_mock(X: InteractionMatrix, target_epoch: int) -> None:

    evaluator = MockEvaluator(X)
    optimizer = MockOptimizer(
        X,
        evaluator,
        fixed_params=dict(target_epoch=target_epoch),
        logger=getLogger("IGNORE"),
    )
    with redirect_stdout(open(os.devnull, "w")):
        with redirect_stderr(open(os.devnull, "w")):
            config, history = optimizer.optimize(n_trials=20, random_seed=42)
    assert len(config) == 2
    assert config["max_epoch"] == target_epoch
    best_index = np.nanargmax(-history.value.values)
    best_target_score_inferred = history.target_score.iloc[best_index]
    best_ndcg = history["ndcg@30"].iloc[best_index]

    assert best_target_score_inferred == pytest.approx(best_ndcg)
    assert config["target_score"] == pytest.approx(best_target_score_inferred)
    assert np.all(best_target_score_inferred >= history.target_score.values)



import os
import sys
from typing import Any, Dict, List

import setuptools
from setuptools import Extension, find_packages, setup
from setuptools.command.build_ext import build_ext

SETUP_DIRECTORY = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(SETUP_DIRECTORY, "Readme.md")) as ifs:
    LONG_DESCRIPTION = ifs.read()

install_requires = (
    [
        "numpy<1.20.0",
        "fastprogress>=0.2",
        "optuna>=2.5.0",
        "pandas>=1.0.0",
        "scikit-learn>=0.21.0",
        "scipy>=1.0",
        "colorlog>=4",
    ],
)

setup_requires = ["pybind11>=2.4", "requests", "setuptools_scm"]
IRSPACK_TESTING = os.environ.get("IRSPACK_TESTING", None) is not None


class get_eigen_include(object):
    EIGEN3_URL = "https://gitlab.com/libeigen/eigen/-/archive/3.3.7/eigen-3.3.7.zip"
    EIGEN3_DIRNAME = "eigen-3.3.7"

    def __str__(self) -> str:
        eigen_include_dir = os.environ.get("EIGEN3_INCLUDE_DIR", None)

        if eigen_include_dir is not None:
            return eigen_include_dir

        SETUP_DIRECTORY = os.path.dirname(__file__)
        target_dir = os.path.join(SETUP_DIRECTORY, self.EIGEN3_DIRNAME)
        if os.path.exists(target_dir):
            return target_dir

        download_target_dir = os.path.join(SETUP_DIRECTORY, "eigen3.zip")
        import zipfile

        import requests

        response = requests.get(self.EIGEN3_URL, stream=True)
        with open(download_target_dir, "wb") as ofs:
            for chunk in response.iter_content(chunk_size=1024):
                ofs.write(chunk)

        with zipfile.ZipFile(download_target_dir) as ifs:
            ifs.extractall()

        return target_dir


class get_pybind_include(object):
    """Helper class to determine the pybind11 include path
    The purpose of this class is to postpone importing pybind11
    until it is actually installed, so that the ``get_include()``
    method can be invoked."""

    def __init__(self, user: Any = False):
        self.user = user

    def __str__(self) -> Any:
        import pybind11

        return pybind11.get_include(self.user)


ext_modules = [
    Extension(
        "irspack.evaluator._core",
        ["cpp_source/evaluator.cpp"],
        include_dirs=[
            get_pybind_include(),
            get_pybind_include(user=True),
            get_eigen_include(),
        ],
        language="c++",
    ),
    Extension(
        "irspack.recommenders._ials",
        ["cpp_source/als/wrapper.cpp"],
        include_dirs=[
            # Path to pybind11 headers
            get_pybind_include(),
            get_pybind_include(user=True),
            get_eigen_include(),
        ],
        language="c++",
    ),
    Extension(
        "irspack.recommenders._knn",
        ["cpp_source/knn/wrapper.cpp"],
        include_dirs=[
            # Path to pybind11 headers
            get_pybind_include(),
            get_pybind_include(user=True),
            get_eigen_include(),
        ],
        language="c++",
    ),
    Extension(
        "irspack.utils._util_cpp",
        ["cpp_source/util.cpp"],
        include_dirs=[
            get_pybind_include(),
            get_pybind_include(user=True),
            get_eigen_include(),
        ],
        language="c++",
    ),
]


# As of Python 3.6, CCompiler has a `has_flag` method.
# cf http://bugs.python.org/issue26689
def has_flag(compiler: Any, flagname: Any) -> bool:
    """Return a boolean indicating whether a flag name is supported on
    the specified compiler.
    """
    import tempfile

    with tempfile.NamedTemporaryFile("w", suffix=".cpp") as f:
        f.write("int main (int argc, char **argv) { return 0; }")
        try:
            compiler.compile([f.name], extra_postargs=[flagname])
        except setuptools.distutils.errors.CompileError:
            return False
    return True


def cpp_flag(compiler: Any) -> str:
    """Return the -std=c++[11/14/17] compiler flag.
    The newer version is prefered over c++11 (when it is available).
    """
    flags = ["-std=c++11"]

    for flag in flags:
        if has_flag(compiler, flag):
            return flag

    raise RuntimeError("Unsupported compiler -- at least C++11 support " "is needed!")


class BuildExt(build_ext):
    """A custom build extension for adding compiler-specific options."""

    if IRSPACK_TESTING:
        c_opts: Dict[str, List[str]] = {
            "msvc": ["/EHsc"],
            "unix": ["-O0", "-coverage", "-g"],
        }
        l_opts: Dict[str, List[str]] = {
            "msvc": [],
            "unix": ["-coverage"],
        }
    else:
        c_opts = {
            "msvc": ["/EHsc"],
            "unix": [],
        }
        l_opts = {
            "msvc": [],
            "unix": [],
        }

    if sys.platform == "darwin":
        darwin_opts = ["-stdlib=libc++", "-mmacosx-version-min=10.7"]
        c_opts["unix"] += darwin_opts
        l_opts["unix"] += darwin_opts

    def build_extensions(self) -> None:
        ct = self.compiler.compiler_type
        opts = self.c_opts.get(ct, [])
        link_opts = self.l_opts.get(ct, [])
        if ct == "unix":
            opts.append('-DVERSION_INFO="%s"' % self.distribution.get_version())
            opts.append(cpp_flag(self.compiler))
            if has_flag(self.compiler, "-fvisibility=hidden"):
                opts.append("-fvisibility=hidden")
        elif ct == "msvc":
            opts.append('/DVERSION_INFO=\\"%s\\"' % self.distribution.get_version())
        for ext in self.extensions:
            ext.extra_compile_args = opts
            ext.extra_link_args = link_opts
        build_ext.build_extensions(self)


def local_scheme(version: Any) -> str:
    return ""


setup(
    name="irspack",
    # version=get_version(),
    url="https://irspack.readthedocs.io/",
    use_scm_version={
        "local_scheme": local_scheme
    },  # https://github.com/pypa/setuptools_scm/issues/342
    author="Tomoki Ohtsuki",
    author_email="tomoki.otsuki129@gmail.com",
    description="Implicit feedback-based recommender system pack",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    ext_modules=ext_modules,
    install_requires=install_requires,
    include_package_data=True,
    setup_requires=setup_requires,
    cmdclass={"build_ext": BuildExt},
    packages=find_packages(),
    python_requires=">=3.6",
)



import re
import time
from logging import Logger
from multiprocessing import Process, Queue
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Tuple, Type
from uuid import uuid1

import numpy as np
import optuna
import pandas as pd
from optuna.samplers import TPESampler
from optuna.storages import RDBStorage
from optuna.trial import TrialState

from irspack.definitions import InteractionMatrix
from irspack.evaluator import Evaluator
from irspack.optimizers.base_optimizer import (
    LowMemoryError,
    get_optimizer_class,
    study_to_dataframe,
)
from irspack.parameter_tuning.parameter_range import Suggestion, is_valid_param_name
from irspack.recommenders import BaseRecommender
from irspack.utils.default_logger import get_default_logger

DEFAULT_SEARCHNAMES = ["RP3beta", "IALS", "DenseSLIM", "AsymmetricCosineKNN", "SLIM"]


def search_one(
    queue: Queue,
    X: InteractionMatrix,
    evaluator: Evaluator,
    optimizer_names: List[str],
    suggest_overwrites: Dict[str, List[Suggestion]],
    intermediate_result_path: Path,
    random_seed: int,
    logger: Logger,
    **kwargs: Any,
) -> None:
    study = optuna.load_study(
        storage=f"sqlite:///{intermediate_result_path.name}",
        study_name="autopilot",
        sampler=TPESampler(seed=random_seed),
    )

    def objective(trial: optuna.Trial) -> float:
        queue.put(trial.number)
        optimizer_name = trial.suggest_categorical("optimizer_name", optimizer_names)
        assert isinstance(optimizer_name, str)
        optimizer = get_optimizer_class(optimizer_name)(
            X,
            evaluator,
            suggest_overwrite=suggest_overwrites[optimizer_name],
            logger=logger,
        )
        result = optimizer.objective_function(optimizer_name + ".")(trial)
        return result

    study.optimize(objective, n_trials=1)


def autopilot(
    X: InteractionMatrix,
    evaluator: Evaluator,
    n_trials: int = 20,
    memory_budget: int = 4000,  # 4GB
    timeout_overall: Optional[int] = None,
    timeout_singlestep: Optional[int] = None,
    algorithms: List[str] = DEFAULT_SEARCHNAMES,
    random_seed: Optional[int] = None,
    logger: Optional[Logger] = None,
    callback: Optional[Callable[[int, pd.DataFrame], None]] = None,
) -> Tuple[Type[BaseRecommender], Dict[str, Any], pd.DataFrame]:
    r"""Given an interaction matrix and an evaluator, search for the best algorithm and its parameters
    (roughly) within the time & space constraints.

    Args:
        X:
            Input interaction matrix.
        evaluator:
            Evaluator to measure the performance of the recommenders.
        n_trials: The maximal number of trials. Defaults to 20.
        memory_budget:
            Optimizers will try search parameters so that memory usage (in megabyte) will not exceed this values.
            An algorithm will not be searched if it inevitably violates this bound.
            Note that this value is quite rough one and will not be respected strictly.
        timeout_overall:
            If set, the total execution time of the trials will not exceed this value (roughly).
        timeout_singlestep:
            If set, a single trial (recommender and a set of its parameter) will not run for more than the value (in seconds).
            Such a trial is considered to have produced  a score value of 0,
            and optuna will avoid suggesting such values (if everything works fine).
            Defaults to `None`.
        algorithms:
            A list of algorithm names to be tried.
            Defaults to `["RP3beta", "IALS", "DenseSLIM", "AsymmetricCosineKNN", "SLIM"]`.
        random_seed:
            The random seed that controls the suggestion behavior.
            Defaults to `None`.
        logger:
            The logger to be used. If `None`, irspack's default logger will be used.
            Defaults to None.
        callback:
            If not `None`, called at the end of every single trial with the following arguments

                1. The current trial's number.
                2. A `pd.DataFrame` that holds history of trial execution.

            Defaults to `None`.

    Raises:
        RuntimeError: If no trials have been completed within given timeout.

    Returns:

        * The best algorithm's recommender class.
        * The best parameters.
        * The dataframe containing the history of trials.
    """
    RNS = np.random.RandomState(random_seed)
    suggest_overwrites: Dict[str, List[Suggestion]] = {}
    optimizer_names: List[str] = []
    db_path = Path(f".autopilot-{uuid1()}.db")
    for rec_name in algorithms:
        optimizer_class_name = rec_name + "Optimizer"
        optimizer_class = get_optimizer_class(optimizer_class_name)
        try:
            suggest_overwrites[
                optimizer_class_name
            ] = optimizer_class.tune_range_given_memory_budget(X, memory_budget)
            optimizer_names.append(optimizer_class_name)
        except LowMemoryError:
            continue

    if not optimizer_names:
        raise RuntimeError("No available algorithm with given memory.")

    if logger is None:
        logger = get_default_logger()

    logger.info("Trying the following algorithms: %s", optimizer_names)
    storage = RDBStorage(
        url=f"sqlite:///{db_path.name}",
    )
    study_id = storage.create_new_study("autopilot")
    start = time.time()
    study = optuna.load_study(
        storage=f"sqlite:///{db_path.name}",
        study_name="autopilot",
    )
    for _ in range(n_trials):

        queue = Queue()  # type: ignore
        p = Process(
            target=search_one,
            args=(
                queue,
                X,
                evaluator,
                optimizer_names,
                suggest_overwrites,
                db_path,
                RNS.randint(0, 2 ** 31),
                logger,
            ),
        )
        p.start()

        process_start = time.time()

        elapsed_at_start = process_start - start
        trial_number: int = queue.get()

        timeout_for_this_process: Optional[int] = None
        if timeout_overall is None:
            timeout_for_this_process = timeout_singlestep
        else:
            timeouf_for_this_process = int(timeout_overall - elapsed_at_start)
            if timeout_singlestep is not None:
                timeout_for_this_process = min(
                    timeouf_for_this_process, timeout_singlestep
                )
        p.join(timeout=timeout_for_this_process)

        if p.exitcode is None:
            logger.info(f"Trial {trial_number} timeout.")
            p.terminate()
            trial_id = storage.get_trial_id_from_study_id_trial_number(
                study_id, trial_number
            )
            try:
                storage.set_trial_values(trial_id, [0.0])
                storage.set_trial_state(trial_id, TrialState.COMPLETE)
            except RuntimeError:  # pragma: no cover
                # this happens if the trial completes before accepting the SIGTERM?
                pass  # pragma: no cover

        if callback is not None:
            callback(trial_number, study_to_dataframe(study))

        now = time.time()
        elapsed = now - start
        if timeout_overall is not None:
            if elapsed > timeout_overall:
                break
    best_params_with_prefix = dict(
        **study.best_trial.params,
        **{
            key: val
            for key, val in study.best_trial.user_attrs.items()
            if is_valid_param_name(key)
        },
    )
    best_params = {
        re.sub(r"^([^\.]*\.)", "", key): value
        for key, value in best_params_with_prefix.items()
    }
    optimizer_name: str = best_params.pop("optimizer_name")
    result_df = study_to_dataframe(study)
    db_path.unlink()
    recommender_class = get_optimizer_class(optimizer_name).recommender_class

    return (recommender_class, best_params, result_df)
